#!/usr/bin/env python
# -*- coding: utf-8 -*-

import dacite
from dataclasses import dataclass, field
from typing import Optional
from elasticsearch import NotFoundError
from nemesis.exceptions import MultipleObjectsReturned
from nemesis.resources import enforce_types, BaseResource
from nemesis.resources.elasticsearch.alias import Alias
from nemesis.schemas.elasticsearch.index_template import IndexTemplateSchema


@enforce_types
@dataclass(frozen=True)
class IndexSettings(BaseResource):
    index: Optional[dict] = None


@enforce_types
@dataclass(frozen=True)
class Template(BaseResource):
    settings: Optional[IndexSettings] = None
    mappings: Optional[dict] = None
    aliases: Optional[Alias] = None


@enforce_types
@dataclass(repr=False, frozen=True)
class IndexTemplate(BaseResource):
    name: str
    index_patterns: list
    template: Template
    version: Optional[int] = None
    priority: Optional[int] = None
    composed_of: Optional[list] = None
    meta: Optional[dict] = None

    @property
    def id(self):
        return self.name

    def asdict(self):
        """
        The "name" field isn't part of the actual body sent to Elasticsearch.
        But it's nice to have on the object we are dealing with.
        """
        d = super().asdict()
        d.pop("name")
        return d

    @classmethod
    def get(cls, client, name):
        """
        Get an index template from Elasticsearch
        """
        try:
            rt = client.indices.get_index_template(name=name, flat_settings=False)
        except NotFoundError:
            return None
        data = rt["index_templates"]
        ret = []

        schema = IndexTemplateSchema()
        for item in data:
            result = schema.load(item)
            template = dacite.from_dict(data_class=cls, data=result)
            ret.append(template)
        if len(ret) > 1:
            raise MultipleObjectsReturned
        else:
            return ret[0]

    def create(self, client):
        try:
            return client.indices.put_index_template(
                name=self.name, body=self.asdict(), create=True
            )
        except Exception as e:
            raise e

    def update(self, client):
        try:
            return client.indices.put_index_template(
                name=self.name, body=self.asdict(), create=False
            )
        except Exception as e:
            raise e

    def delete(self, client):
        try:
            return client.indices.delete_index_template(name=self.name)
        except Exception as e:
            raise e



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from mock import MagicMock
from nemesis import Nemesis
from elasticsearch import Elasticsearch
from nemesis.resources.elasticsearch.index_template import Template


CLOUD_ID = "not-a-real-deployment:ZXUtd2VzdC0xLmF3cy5mb3VuZC5pbyRlOTY1NjUwNzU4Y2Y0ODAzYjZlNWJkNWZkNDZjNzI3ZCQzMGY5OGQyZTMwOGI0OGUwODNhM2I2ODEwMGU1MTU0MQ=="
ES_HOST = "http://localhost:9200"
USERNAME = "elastic"
PASSWORD = "changeme"

def test_init_nemesis_cloud_id():
    n = Nemesis(username=USERNAME, password=PASSWORD, cloud_id=CLOUD_ID)
    assert isinstance(n.client, Elasticsearch)


def test_init_nemesis_url():
    n = Nemesis(username=USERNAME, password=PASSWORD, es_host=ES_HOST)
    assert isinstance(n.client, Elasticsearch)


def test_get_client_for_resource():
    template = Template()
    n = Nemesis(username=USERNAME, password=PASSWORD, cloud_id=CLOUD_ID)
    assert n.get_client_for(template) == n.client


def test_register_resource():
    resource = MagicMock()
    resource.get.return_value = Template()
    n = Nemesis(username=USERNAME, password=PASSWORD, es_host=ES_HOST)
    n.register(resource)
    assert len(n.resources) == 1
    r = n.resources[0]
    assert r["force"] is False
    assert r["post_deploy"] is None
    assert r["pre_deploy"] is None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional
from dataclasses import dataclass
from typing import Optional, Dict, cast

from nemesis.resources import enforce_types, BaseResource


@enforce_types
@dataclass(frozen=True)
class QueryDSL(BaseResource):
    id: Optional[str] = None
    bool: Optional[dict] = None
    boolstring: Optional[dict] = None
    common: Optional[dict] = None
    constant_score: Optional[dict] = None
    custom_filters_score: Optional[dict] = None
    dis_max: Optional[dict] = None
    distance_feature: Optional[dict] = None
    exists: Optional[dict] = None
    field: Optional[dict] = None
    function_score: Optional[dict] = None
    fuzzy: Optional[dict] = None
    geo_shape: Optional[dict] = None
    has_child: Optional[dict] = None
    has_parent: Optional[dict] = None
    ids: Optional[dict] = None
    indices: Optional[dict] = None
    match: Optional[dict] = None
    match_all: Optional[dict] = None
    match_phrase: Optional[dict] = None
    match_phrase_prefix: Optional[dict] = None
    nested: Optional[dict] = None
    percolate: Optional[dict] = None
    prefix: Optional[dict] = None
    query_string: Optional[dict] = None
    range: Optional[dict] = None
    regexp: Optional[dict] = None
    script: Optional[dict] = None
    simple_query_string: Optional[dict] = None
    span_containing: Optional[dict] = None
    span_first: Optional[dict] = None
    span_multi: Optional[dict] = None
    span_near: Optional[dict] = None
    span_not: Optional[dict] = None
    span_or: Optional[dict] = None
    span_term: Optional[dict] = None
    span_within: Optional[dict] = None
    term: Optional[dict] = None
    wildcard: Optional[dict] = None
    wrapper: Optional[dict] = None

    def asdict(self):
        """
        The "id" field isn't part of the actual body sent to Elasticsearch.
        But it's nice to have on the object we are dealing with.
        """
        d = super().asdict()
        d.pop("id")
        return d



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from marshmallow import Schema, fields, pre_load


class LogstashPipelineSchema(Schema):
    id = fields.Str()
    description = fields.Str()
    username = fields.Str()
    last_modified = fields.DateTime()
    pipeline = fields.Str()
    pipeline_metadata = fields.Dict()
    pipeline_settings = fields.Dict()

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many, **kwargs):
        keys = list(data.keys())
        pipeline_id = keys[0]
        data = data.pop(pipeline_id)
        data["id"] = pipeline_id
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from dataclasses import FrozenInstanceError
from nemesis.resources.elasticsearch.index_template import (
    IndexTemplate,
    Template,
    IndexSettings,
)
from elasticsearch import Elasticsearch, RequestError, NotFoundError
from nemesis.exceptions import MultipleObjectsReturned
from nemesis.tests import es_client, nemesis_client


@pytest.fixture
def template(es_client):
    template = IndexTemplate(
        name="test-template",
        index_patterns=["test-foo"],
        template=Template(
            settings=IndexSettings(index={"number_of_replicas": "2"}),
        ),
    )
    yield template
    # clear remote resources if they exist
    try:
        result = template.delete(es_client)
        assert result == {"acknowledged": True}
    except NotFoundError:
        pass


@pytest.fixture
def template2(es_client):
    template = IndexTemplate(
        name="test-template-2",
        index_patterns=["test-bar"],
        template=Template(settings=IndexSettings(index={"number_of_replicas": 2})),
    )
    yield template
    # clear remote resources if they exist
    try:
        result = template.delete(es_client)
        assert result == {"acknowledged": True}
    except NotFoundError:
        pass


def test_asdict(template):
    """
    asdict method pop's the `name` field from the object to match
    how it would be sent or received from Elasticsearch
    """

    assert template.asdict() == {
        "index_patterns": ["test-foo"],
        "template": {
            "settings": {"index": {"number_of_replicas": "2"}},
        },
    }


def test_frozen_object(template):
    with pytest.raises(FrozenInstanceError):
        template.index_patterns = ["raise-error"]


def test_id(template):
    assert template.id == "test-template"


def test_get_multiple(template, template2, es_client):
    result = template.create(es_client)
    assert result == {"acknowledged": True}

    result = template2.create(es_client)
    assert result == {"acknowledged": True}

    with pytest.raises(MultipleObjectsReturned):
        IndexTemplate.get(es_client, "test-template*")


def test_create_get_update_delete(template, es_client):
    assert IndexTemplate.get(es_client, template.id) is None

    result = template.create(es_client)
    assert result == {"acknowledged": True}

    result = IndexTemplate.get(es_client, template.id)
    assert result.asdict() == template.asdict()

    with pytest.raises(RequestError):
        result = template.create(es_client)

    template.index_patterns.append("foo-*")
    result = template.update(es_client)
    assert result == {"acknowledged": True}

    result = template.delete(es_client)
    assert result == {"acknowledged": True}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
from marshmallow import Schema, fields, pre_load
from nemesis.schemas.elasticsearch.querydsl import QueryDSLSchema


class IngestPipelineSchema(Schema):
    processors = fields.List(fields.Dict())
    description = fields.Str()
    on_failure = fields.List(fields.Dict())
    version = fields.Int()
    meta = fields.Dict(data_key="_meta")
    id = fields.Str()

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many, **kwargs):
        keys = list(data.keys())
        pipeline_id = keys[0]
        data = data.pop(pipeline_id)
        data["id"] = pipeline_id
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from deepdiff.helper import NotPresent

ADD = "[add]+[/]"
REMOVE = "[remove]-[/]"
CHANGE = "[change]~[/]"


class Diff:
    local_resource = None
    remote_resource = None
    status = "update"

    def __init__(self, local_resource, remote_resource):
        self.local_resource = local_resource
        self.remote_resource = remote_resource
        self.deltas = []
        self.do_diff()

    def _values_changed(self, diff):
        self.status = "update"
        for item in diff:
            path = item.path().replace("root", "")
            self.deltas.append(
                {
                    "path": path,
                    "long": f"{ADD} {json.dumps(item.t1, indent=2)}\n{REMOVE} {json.dumps(item.t2, indent=2)}",
                    "short": f"{CHANGE} {path}",
                }
            )

    def _dictionary_item_added(self, diff):
        self.status = "update"
        for item in diff:
            path = item.path().replace("root", "")
            if isinstance(item.t1, NotPresent):
                # This scenario means the object was edited in Kibana and the resource defined
                # in nemesis doesn't match what's remote.
                # We have an item that doesn't exist locally, yet exists remotely.
                # This can happen when a logstash pipeline is edited in Kibana, and changes the `metadata`.
                self.deltas.append(
                    {
                        "path": path,
                        "long": f"{ADD} {json.dumps(item.t2, indent=2)}",
                        "short": f"{ADD} {path}",
                    }
                )
            else:
                self.deltas.append(
                    {
                        "path": path,
                        "long": f"{ADD} {json.dumps(item.t2, indent=2)}",
                        "short": f"{ADD} {path}",
                    }
                )

    def _dictionary_item_removed(self, diff):
        if self.remote_resource is None:
            self.status = "create"

        for item in diff:
            path = item.path().replace("root", "")
            if isinstance(item.t2, NotPresent):
                # If the remote_resource is None, and item.t2 is NotPresent
                # This means the remote object does Not exist, and we are adding
                # the full resource.
                self.deltas.append(
                    {
                        "path": path,
                        "long": f"{ADD} {json.dumps(item.t1, indent=2)}",
                        "short": f"{ADD} {path}",
                    }
                )
            else:
                self.deltas.append(
                    {
                        "path": path,
                        "long": f"{REMOVE} {json.dumps(item.t1, indent=2)}",
                        "short": f"{REMOVE} {path}",
                    }
                )

    def _iterable_item_added(self, diff):
        self.status = "update"
        for item in diff:
            path = item.path().replace("root", "")
            self.deltas.append(
                {
                    "path": path,
                    "long": f"{ADD} {json.dumps(item.t2, indent=2)}",
                    "short": f"{ADD} {path}",
                }
            )

    def _iterable_item_removed(self, diff):
        self.status = "update"
        for item in diff:
            path = item.path().replace("root", "")
            self.deltas.append(
                {
                    "path": path,
                    "long": f"{REMOVE} {json.dumps(item.t1, indent=2)}",
                    "short": f"{REMOVE} {path}",
                }
            )

    def do_diff(self):
        deepdiff = self.local_resource.diff(self.remote_resource)
        if not deepdiff:
            self.status = "unchanged"
        for reason, diff in deepdiff.items():
            if reason == "values_changed":
                self._values_changed(diff)
            elif reason == "dictionary_item_added":
                self._dictionary_item_added(diff)
            elif reason == "dictionary_item_removed":
                self._dictionary_item_removed(diff)
            elif reason == "iterable_item_removed":
                self._iterable_item_removed(diff)
            elif reason == "iterable_item_added":
                self._iterable_item_added(diff)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import dacite
from elasticsearch import NotFoundError
from dataclasses import dataclass, field
from typing import Optional
from nemesis.schemas.elasticsearch.ingest_pipeline import IngestPipelineSchema
from nemesis.resources import enforce_types, BaseResource


@enforce_types
@dataclass(repr=False, frozen=True)
class IngestPipeline(BaseResource):
    id: str
    processors: list
    description: Optional[str] = None
    on_failure: Optional[list] = None
    version: Optional[int] = None
    _meta: Optional[dict] = None

    @classmethod
    def get(cls, client, pipeline_id):
        """
        Get an ingest pipeline from Elasticsearch
        """
        try:
            pipeline = client.ingest.get_pipeline(id=pipeline_id)
        except NotFoundError:
            return None
        schema = IngestPipelineSchema()
        result = schema.load(pipeline)
        pipeline = dacite.from_dict(data_class=cls, data=result)
        return pipeline

    def create(self, client):
        body = self.asdict()
        body.pop("id")
        return client.ingest.put_pipeline(id=self.id, body=body)

    def delete(self, client):
        try:
            return client.ingest.delete_pipeline(id=self.id)
        except Exception as e:
            raise e

    def update(self, client):
        return self.create(client)

    def simulate(self, client, docs):
        pipeline = self.asdict()
        pipeline.pop("id")
        body = {"pipeline": pipeline, "docs": docs}
        ret = client.ingest.simulate(body, verbose=True)
        return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from dataclasses import FrozenInstanceError
from nemesis.resources.elasticsearch.index import Index
from nemesis.resources.elasticsearch.ingest_pipeline import IngestPipeline
from nemesis.resources.elasticsearch.transform import (
    Transform,
    Source,
    QueryDSL,
    Dest,
    Sync,
    SyncTime,
    Pivot,
    RetentionPolicy,
    RetentionPolicyTime,
)
from elasticsearch import Elasticsearch, RequestError, NotFoundError
from nemesis.tests.resources.elasticsearch.test_ingest_pipeline import pipeline
from nemesis.tests import es_client, nemesis_client
from nemesis.schemas.elasticsearch.transform import TransformSchema


@pytest.fixture
def dest_index(es_client):
    index = Index(name="kibana_sample_data_ecommerce_transform1")
    yield index

    # clear remote resources if they exist
    try:
        result = index.delete(es_client)
        assert result == {"acknowledged": True}
    except NotFoundError:
        pass


@pytest.fixture
def source_index(es_client):
    index = Index(name="kibana_sample_data_ecommerce")
    yield index

    # clear remote resources if they exist
    try:
        result = index.delete(es_client)
        assert result == {"acknowledged": True}
    except NotFoundError:
        pass


@pytest.fixture
def transform(es_client, dest_index, source_index, pipeline):
    transform = Transform(
        id="test-transform",
        description="Maximum priced ecommerce data by customer_id in Asia",
        retention_policy=RetentionPolicy(
            time=RetentionPolicyTime(field="order_date", max_age="30d")
        ),
        dest=Dest(index=dest_index.id, pipeline=pipeline.id),
        pivot=Pivot(
            group_by={"customer_id": {"terms": {"field": "customer_id"}}},
            aggregations={"max_price": {"max": {"field": "taxful_total_price"}}},
        ),
        sync=Sync(
            time=SyncTime(field="order_date", delay="60s"),
        ),
        frequency="5m",
        source=Source(
            index=[
                source_index.id,
            ],
            query=QueryDSL(term={"geoip.continent_name": {"value": "Asia"}}),
        ),
    )
    yield transform

    # clear remote resources if they exist
    try:
        result = transform.delete(es_client)
        assert result == {"acknowledged": True}
    except NotFoundError:
        pass


def test_latest_or_pivot(source_index, dest_index, pipeline):
    with pytest.raises(TypeError):
        Transform(
            id="test-transform",
            description="Maximum priced ecommerce data by customer_id in Asia",
            retention_policy=RetentionPolicy(
                time=RetentionPolicyTime(field="order_date", max_age="30d")
            ),
            dest=Dest(index=dest_index.id, pipeline=pipeline.id),
            sync=Sync(
                time=SyncTime(field="order_date", delay="60s"),
            ),
            frequency="5m",
            source=Source(
                index=[
                    source_index.id,
                ],
                query=QueryDSL(term={"geoip.continent_name": {"value": "Asia"}}),
            ),
        )


def test_asdict(transform):
    """ """
    assert transform.asdict() == {
        "source": {
            "index": ["kibana_sample_data_ecommerce"],
            "query": {"term": {"geoip.continent_name": {"value": "Asia"}}},
        },
        "dest": {
            "index": "kibana_sample_data_ecommerce_transform1",
            "pipeline": "test-pipeline",
        },
        "id": "test-transform",
        "pivot": {
            "aggregations": {"max_price": {"max": {"field": "taxful_total_price"}}},
            "group_by": {"customer_id": {"terms": {"field": "customer_id"}}},
        },
        "sync": {"time": {"field": "order_date", "delay": "60s"}},
        "retention_policy": {"time": {"field": "order_date", "max_age": "30d"}},
        "description": "Maximum priced ecommerce data by customer_id in Asia",
        "frequency": "5m",
    }


def test_frozen_object(transform):
    with pytest.raises(FrozenInstanceError):
        transform.processors = [{}]


def test_id(transform):
    assert transform.id == "test-transform"


def test_get_not_found(es_client):
    assert Transform.get(es_client, "notfound") is None


def test_create_multiple(es_client, transform, source_index, dest_index, pipeline):
    result = dest_index.create(es_client)
    assert result == {
        "acknowledged": True,
        "shards_acknowledged": True,
        "index": "kibana_sample_data_ecommerce_transform1",
    }
    result = source_index.create(es_client)
    assert result == {
        "acknowledged": True,
        "shards_acknowledged": True,
        "index": "kibana_sample_data_ecommerce",
    }

    result = pipeline.create(es_client)
    assert result == {"acknowledged": True}

    result = transform.create(es_client)
    assert result == {"acknowledged": True}

    with pytest.raises(RequestError):
        transform.create(es_client)


def test_create_update_delete(transform, dest_index, source_index, pipeline, es_client):
    result = dest_index.create(es_client)
    assert result == {
        "acknowledged": True,
        "shards_acknowledged": True,
        "index": "kibana_sample_data_ecommerce_transform1",
    }
    result = source_index.create(es_client)
    assert result == {
        "acknowledged": True,
        "shards_acknowledged": True,
        "index": "kibana_sample_data_ecommerce",
    }

    result = pipeline.create(es_client)
    assert result == {"acknowledged": True}

    result = transform.create(es_client)
    assert result == {"acknowledged": True}

    result = Transform.get(es_client, transform.id)
    assert result == transform

    t = transform.asdict()
    t["description"] = "change me"
    new_transform = Transform.fromdict(schemaclass=TransformSchema, body=t)

    with pytest.raises(RequestError):
        new_transform.update(es_client)

    result = transform.delete(es_client)
    assert result == {"acknowledged": True}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from dataclasses import FrozenInstanceError
from nemesis.resources.elasticsearch.index import (
    Index,
    IndexSettings,
)
from elasticsearch import Elasticsearch, RequestError, NotFoundError
from nemesis.exceptions import MultipleObjectsReturned
from nemesis.tests import es_client, nemesis_client


@pytest.fixture
def index(es_client):
    indexsettings = IndexSettings(
        index={
            "routing": {
                "allocation": {"include": {"_tier_preference": "data_content"}}
            },
            "number_of_shards": "1",
            "number_of_replicas": "2",
        }
    )
    index = Index(name="test-index", settings=indexsettings)
    yield index
    # clear remote resources if they exist
    try:
        result = index.delete(es_client)
        assert result == {"acknowledged": True}
    except NotFoundError:
        pass


def test_asdict(index):
    """
    asdict method pop's the `name` field from the object to match
    how it would be sent or received from Elasticsearch
    """
    assert index.asdict() == {
        "settings": {
            "index": {
                "routing": {
                    "allocation": {"include": {"_tier_preference": "data_content"}}
                },
                "number_of_shards": "1",
                "number_of_replicas": "2",
            }
        }
    }


def test_id(index):
    assert index.id == "test-index"


def test_create_get_update_delete(index, es_client):
    assert Index.get(es_client, index.id) is None

    result = index.create(es_client)
    assert result == {
        "acknowledged": True,
        "index": "test-index",
        "shards_acknowledged": True,
    }

    result = Index.get(es_client, index.id)
    assert result.asdict() == index.asdict()

    with pytest.raises(RequestError):
        result = index.create(es_client)

    result = index.delete(es_client)
    assert result == {"acknowledged": True}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from dataclasses import FrozenInstanceError
from nemesis.resources.elasticsearch.querydsl import QueryDSL
from elasticsearch import Elasticsearch, RequestError

from nemesis.tests import es_client, nemesis_client


@pytest.fixture
def query():
    return QueryDSL(id="test-query", bool={"must": {"term": {"user.id": "kimchy"}}})


def test_asdict(query):
    """
    asdict method pop's the `name` field from the object to match
    how it would be sent or received from Elasticsearch
    """

    assert query.asdict() == {"bool": {"must": {"term": {"user.id": "kimchy"}}}}


def test_frozen_object(query):
    with pytest.raises(FrozenInstanceError):
        query.bool = {"must": {"term": {"user.id": "fxdgear"}}}


def test_id(query):
    assert query.id == "test-query"


def test_create_update_delete(query, es_client):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from mock import MagicMock
from deepdiff import DeepDiff

from nemesis.utils import Diff
from nemesis.resources.elasticsearch.index_template import IndexTemplate, Template


def test_iterable_item_added():
    t1 = IndexTemplate(
        name="name",
        index_patterns=[
            "a",
        ],
        template=Template(),
    )
    t2 = IndexTemplate(
        name="name",
        index_patterns=["a", "b"],
        template=Template(),
    )
    diff = Diff(t1, t2)
    assert len(diff.deltas) == 1
    delta = diff.deltas[0]
    assert delta["long"] == '[add]+[/] "b"'
    assert delta["short"] == "[add]+[/] ['index_patterns'][1]"
    assert delta["path"] == "['index_patterns'][1]"


def test_iterable_item_added_remote_dne():
    t1 = IndexTemplate(
        name="name",
        index_patterns=[
            "a",
        ],
        template=Template(),
    )
    diff = Diff(t1, None)
    assert len(diff.deltas) == 2

    delta1 = diff.deltas[0]
    delta2 = diff.deltas[1]
    assert delta1["long"] == '[add]+[/] [\n  "a"\n]'
    assert delta1["short"] == "[add]+[/] ['index_patterns']"
    assert delta1["path"] == "['index_patterns']"
    assert delta2["long"] == "[add]+[/] {}"
    assert delta2["short"] == "[add]+[/] ['template']"
    assert delta2["path"] == "['template']"


def test_iterable_item_removed():
    t1 = IndexTemplate(
        name="name",
        index_patterns=["a", "b"],
        template=Template(),
    )
    t2 = IndexTemplate(
        name="name",
        index_patterns=[
            "a",
        ],
        template=Template(),
    )
    diff = Diff(t1, t2)
    assert len(diff.deltas) == 1
    delta = diff.deltas[0]
    assert delta["long"] == '[remove]-[/] "b"'
    assert delta["short"] == "[remove]-[/] ['index_patterns'][1]"
    assert delta["path"] == "['index_patterns'][1]"


def test_dictionary_item_added():
    t1 = Template(
        mappings={"a": {"b": "c"}},
    )
    t2 = Template(
        mappings={"a": {"b": "c", "d": "e"}},
    )
    diff = Diff(t1, t2)
    assert len(diff.deltas) == 1
    delta = diff.deltas[0]
    assert delta["long"] == '[add]+[/] "e"'
    assert delta["short"] == "[add]+[/] ['mappings']['a']['d']"
    assert delta["path"] == "['mappings']['a']['d']"


def test_dictionary_item_added_remote_dne():
    t1 = Template(
        mappings={"a": {"b": "c"}},
    )
    diff = Diff(t1, None)
    assert len(diff.deltas) == 1
    delta = diff.deltas[0]
    assert delta["long"] == '[add]+[/] {\n  "a": {\n    "b": "c"\n  }\n}'
    assert delta["short"] == "[add]+[/] ['mappings']"
    assert delta["path"] == "['mappings']"


def test_values_changed():
    t1 = Template(
        mappings={"a": {"b": "c"}},
    )
    t2 = Template(
        mappings={"a": {"b": "d"}},
    )
    diff = Diff(t1, t2)
    assert len(diff.deltas) == 1
    delta = diff.deltas[0]
    assert delta["long"] == '[add]+[/] "c"\n[remove]-[/] "d"'
    assert delta["short"] == "[change]~[/] ['mappings']['a']['b']"
    assert delta["path"] == "['mappings']['a']['b']"




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import calendar
import json

import dacite
from elasticsearch import RequestError, NotFoundError
from typing import Optional
from dataclasses import dataclass, field, asdict
from nemesis.resources import enforce_types, BaseResource
from nemesis.resources.elasticsearch.querydsl import QueryDSL
from nemesis.schemas.elasticsearch.transform import TransformSchema
from nemesis.resources.elasticsearch.index import Index
from nemesis.resources.elasticsearch.ingest_pipeline import IngestPipeline


@enforce_types
@dataclass(frozen=True)
class Dest(BaseResource):
    index: str
    pipeline: Optional[str] = None


@enforce_types
@dataclass(frozen=True)
class Latest(BaseResource):
    sort: str
    unique_key: list[str]


@enforce_types
@dataclass(frozen=True)
class Pivot(BaseResource):
    aggregations: dict
    group_by: dict


@enforce_types
@dataclass(frozen=True)
class RetentionPolicyTime(BaseResource):
    field: str
    max_age: str


@enforce_types
@dataclass(frozen=True)
class RetentionPolicy(BaseResource):
    time: RetentionPolicyTime


@enforce_types
@dataclass(frozen=True)
class Settings(BaseResource):
    docs_per_second: Optional[float]
    dates_as_epoch_millis: Optional[bool]
    align_checkpoints: Optional[bool]
    max_page_search_size: Optional[int]


@enforce_types
@dataclass(frozen=True)
class SyncTime(BaseResource):
    field: str
    delay: str = "60s"


@enforce_types
@dataclass(frozen=True)
class Sync(BaseResource):
    time: SyncTime


@enforce_types
@dataclass(frozen=True)
class Source(BaseResource):
    """
    Source is a required parameter of Transform
    https://www.elastic.co/guide/en/elasticsearch/reference/current/put-transform.html#put-transform-request-body
    """

    index: list
    runtime_mappings: Optional[dict] = None
    query: Optional[QueryDSL] = None


@dataclass(repr=False, frozen=True)
class Transform(BaseResource):
    """
    https://www.elastic.co/guide/en/elasticsearch/reference/current/put-transform.html#put-transform-request-body
    """

    source: Source
    dest: Dest
    id: Optional[str] = None
    pivot: Optional[Pivot] = None
    latest: Optional[Latest] = None
    sync: Optional[Sync] = None
    retention_policy: Optional[RetentionPolicy] = None
    settings: Optional[Settings] = None
    description: Optional[str] = None
    frequency: str = "1m"

    def __post_init__(self):
        if self.pivot is None and self.latest is None:
            raise TypeError("Value needed for `latest` or `pivot` field")

    @classmethod
    def get(cls, client, transform_id):
        """
        Get a transform from Elasticsearch
        """
        try:
            rt = client.transform.get_transform(transform_id=transform_id)
        except NotFoundError:
            return None
        transforms = rt["transforms"]
        ret = []
        for transform in transforms:
            ret.append(cls.fromdict(schemaclass=TransformSchema, body=transform))
        if len(ret) > 1:
            return ret
        else:
            return ret[0]

    def create(self, client, defer_validation=False, *args, **kwargs):
        try:
            ret = client.transform.put_transform(
                transform_id=self.id,
                body=self.asdict(),
                defer_validation=defer_validation,
                request_timeout=90,
                *args,
                **kwargs,
            )
        except RequestError as e:
            raise e
        return ret

    def delete(self, client, force=False, *args, **kwargs):
        try:
            return client.transform.delete_transform(
                transform_id=self.id, force=force, *args, **kwargs
            )
        except Exception as e:
            raise e

    def stop(
        self,
        client,
        allow_no_match=True,
        force=False,
        timeout="30s",
        wait_for_checkpoint=False,
        wait_for_completion=False,
        *args,
        **kwargs,
    ):
        try:
            return client.transform.stop_transform(
                transform_id=self.id,
                allow_no_match=allow_no_match,
                force=force,
                timeout=timeout,
                wait_for_checkpoint=wait_for_checkpoint,
                wait_for_completion=wait_for_completion,
                *args,
                **kwargs,
            )
        except Exception as e:
            raise e

    def start(self, client, timeout="30s", *args, **kwargs):
        try:
            return client.transform.start_transform(
                transform_id=self.id, timeout=timeout, *args, **kwargs
            )
        except Exception as e:
            raise e

    def update(self, client, *args, **kwargs):
        return self.create(client)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from rich.console import Console
from rich.text import Text
from rich.theme import Theme

nemesis_theme = Theme(
    {
        "resource": "bold purple",
        "add": "bold green",
        "change": "yellow",
        "remove": "bold red",
        "error": "red",
        "function": "blue",
    }
)
console = Console(theme=nemesis_theme)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from marshmallow import Schema, fields, pre_load
from nemesis.schemas.elasticsearch.querydsl import QueryDSLSchema
from nemesis.schemas.elasticsearch.alias import AliasSchema


class IndexSettingsSchema(Schema):
    index = fields.Dict()

    @pre_load(pass_many=True)
    def remove_ignored_fields(self, data, many, **kwargs):
        ignored_fields = ["creation_date", "provided_name", "uuid", "version"]
        d = data["index"]

        for field in ignored_fields:
            if field in d.keys():
                d.pop(field)
        return data


class IndexSchema(Schema):
    name = fields.Str()
    aliases = fields.Nested(AliasSchema, data_key="aliases", allow_none=True)
    settings = fields.Nested(IndexSettingsSchema, data_key="settings", allow_none=True)
    mappings = fields.Dict(data_key="mappings", allow_none=True)

    def remove_empty_values(self, data):
        for k, v in data.items():
            if not v:
                data[k] = None
        return data

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many, **kwargs):
        keys = list(data.keys())
        id = keys[0]
        data = data.pop(id)
        data["name"] = id
        data = self.remove_empty_values(data)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from dataclasses import FrozenInstanceError
from nemesis.resources.elasticsearch.watcher import (
    Watch,
    Trigger,
    Input,
    Search,
    SearchRequest,
    Body,
    QueryDSL,
    Condition,
)

from elasticsearch import Elasticsearch, RequestError, NotFoundError
from nemesis.exceptions import MultipleObjectsReturned
from nemesis.tests import es_client, nemesis_client


@pytest.fixture
def watch(es_client):
    watch = Watch(
        watch_id="test-watch",
        trigger=Trigger(schedule={"cron": "0 0/1 * * * ?"}),
        input=Input(
            search=Search(
                request=SearchRequest(
                    indices=[
                        "logstash*",
                    ],
                    body=Body(
                        query=QueryDSL(
                            bool={
                                "must": {"match": {"response": 404}},
                                "filter": {
                                    "range": {
                                        "@timestamp": {
                                            "from": "{{ctx.trigger.scheduled_time}}||-5m",
                                            "to": "{{ctx.trigger.triggered_time}}",
                                        }
                                    }
                                },
                            }
                        )
                    ),
                )
            )
        ),
        condition=Condition(compare={"ctx.payload.hits.total": {"gt": 0}}),
        actions={
            "email_admin": {
                "email": {
                    "profile": "standard",
                    "to": [
                        "admin@domain.host.com",
                    ],
                    "subject": "404 recently encountered",
                }
            }
        },
    )
    yield watch
    # clear remote resources if they exist
    try:
        result = watch.delete(es_client)
        assert result["found"]
    except NotFoundError:
        pass


def test_asdict(watch):
    """
    asdict method pop's the `name` field from the object to match
    how it would be sent or received from Elasticsearch
    """
    assert watch.asdict() == {
        "trigger": {"schedule": {"cron": "0 0/1 * * * ?"}},
        "input": {
            "search": {
                "request": {
                    "indices": ["logstash*"],
                    "body": {
                        "query": {
                            "bool": {
                                "must": {"match": {"response": 404}},
                                "filter": {
                                    "range": {
                                        "@timestamp": {
                                            "from": "{{ctx.trigger.scheduled_time}}||-5m",
                                            "to": "{{ctx.trigger.triggered_time}}",
                                        }
                                    }
                                },
                            }
                        }
                    },
                }
            }
        },
        "condition": {"compare": {"ctx.payload.hits.total": {"gt": 0}}},
        "actions": {
            "email_admin": {
                "email": {
                    "profile": "standard",
                    "to": [
                        "admin@domain.host.com",
                    ],
                    "subject": "404 recently encountered",
                }
            }
        },
    }


def test_frozen_object(watch):
    with pytest.raises(FrozenInstanceError):
        watch.index_patterns = ["raise-error"]


def test_id(watch):
    assert watch.id == "test-watch"


def test_create_get_update_delete(watch, es_client):
    assert Watch.get(es_client, watch.id) is None

    result = watch.create(es_client)
    assert result["created"]

    result = Watch.get(es_client, watch.id)
    assert result.asdict() == watch.asdict()

    result = watch.create(es_client)
    assert result["created"] is False

    result = watch.delete(es_client)
    assert result["found"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from marshmallow import Schema, fields


class QueryDSLSchema(Schema):
    bool = fields.Dict()
    boolstring = fields.Dict()
    common = fields.Dict()
    constant_score = fields.Dict()
    custom_filters_score = fields.Dict()
    dis_max = fields.Dict()
    distance_feature = fields.Dict()
    exists = fields.Dict()
    field = fields.Dict()
    function_score = fields.Dict()
    fuzzy = fields.Dict()
    geo_shape = fields.Dict()
    has_child = fields.Dict()
    has_parent = fields.Dict()
    ids = fields.Dict()
    indices = fields.Dict()
    match = fields.Dict()
    match_all = fields.Dict()
    match_phrase = fields.Dict()
    match_phrase_prefix = fields.Dict()
    nested = fields.Dict()
    percolate = fields.Dict()
    prefix = fields.Dict()
    query_string = fields.Dict()
    range = fields.Dict()
    regexp = fields.Dict()
    script = fields.Dict()
    simple_query_string = fields.Dict()
    span_containing = fields.Dict()
    span_first = fields.Dict()
    span_multi = fields.Dict()
    span_near = fields.Dict()
    span_not = fields.Dict()
    span_or = fields.Dict()
    span_term = fields.Dict()
    span_within = fields.Dict()
    term = fields.Dict()
    wildcard = fields.Dict()
    wrapper = fields.Dict()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import os

from nemesis import Nemesis

TEST_USERNAME = os.environ.get('TEST_USERNAME', 'elastic')
TEST_PASSWORD = os.environ.get('TEST_PASSWORD', 'changeme')
TEST_HOST = os.environ.get('TEST_HOST', 'http://localhost:9200')


def get_client(username=TEST_USERNAME, password=TEST_PASSWORD, es_host=TEST_HOST):
    return Nemesis(username=username, password=password, es_host=es_host)

@pytest.fixture
def nemesis_client():
    return get_client()

@pytest.fixture
def es_client():
    n = get_client()
    return n.client



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from marshmallow import Schema, fields, pre_load, EXCLUDE
from nemesis.schemas.elasticsearch.querydsl import QueryDSLSchema


class ApplicationSchema(Schema):
    application = fields.Str()
    privileges = fields.List(fields.Str())
    resources = fields.List(fields.Str())


class IndexSchema(Schema):
    field_security = fields.Dict()
    names = fields.List(fields.Str())
    privileges = fields.List(fields.Str())
    query = fields.Nested(QueryDSLSchema, data_key="query")
    allow_restricted_indices = fields.Bool()


class RoleSchema(Schema):
    name = fields.Str()
    applications = fields.List(
        fields.Nested(ApplicationSchema, data_key="applications")
    )
    cluster = fields.List(fields.Str())
    _global = fields.Dict(data_key="global")
    indices = fields.List(fields.Nested(IndexSchema, data_key="indices"))
    metadata = fields.Dict()
    run_as = fields.List(fields.Str())

    class Meta:
        unknown = EXCLUDE

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many, **kwargs):
        keys = list(data.keys())
        role_name = keys[0]
        data = data.pop(role_name)
        data["name"] = role_name
        for c, index in enumerate(data["indices"]):
            if isinstance(index["query"], str):
                index["query"] = json.loads(index["query"])
        return data


class RoleMappingSchema(Schema):
    name = fields.Str()
    enabled = fields.Boolean()
    roles = fields.List(fields.Str())
    role_templates = fields.List(fields.Dict())
    rules = fields.Dict()
    metadata = fields.Dict()

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many, **kwargs):
        keys = list(data.keys())
        name = keys[0]
        data = data.pop(name)
        data["name"] = name
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import dacite
from marshmallow.exceptions import ValidationError
from dataclasses import dataclass, field
from typing import Optional, Union
from nemesis.resources import enforce_types, BaseResource
from nemesis.schemas.elasticsearch.watcher import WatchSchema
from nemesis.resources.elasticsearch.querydsl import QueryDSL
from elasticsearch import NotFoundError


@enforce_types
@dataclass(repr=False, frozen=True)
class Trigger(BaseResource):
    schedule: dict


@enforce_types
@dataclass(repr=False, frozen=True)
class Body(BaseResource):
    query: QueryDSL
    size: Optional[int] = None
    sort: Optional[dict] = None


@enforce_types
@dataclass(repr=False, frozen=True)
class SearchTemplate(BaseResource):
    id: str
    params: dict


@enforce_types
@dataclass(repr=False, frozen=True)
class SearchRequest(BaseResource):
    indices: list
    body: Body
    template: Optional[SearchTemplate] = None


@enforce_types
@dataclass(repr=False, frozen=True)
class Search(BaseResource):
    request: SearchRequest
    extract: Optional[list] = None


@enforce_types
@dataclass(repr=False, frozen=True)
class HttpRequest(BaseResource):
    scheme: Optional[str] = "http"
    host: Optional[str] = None
    port: Optional[int] = None
    path: Optional[str] = None
    url: Optional[str] = None
    method: Optional[str] = "get"
    body: Optional[str] = None
    params: Optional[dict] = None
    headers: Optional[dict] = None
    auth: Optional[dict] = None
    proxy: Optional[dict] = None
    connection_timeout: Optional[str] = "10s"
    read_timeout: Optional[str] = "10s"
    extract: Optional[list] = None
    response_content_type: Optional[str] = "json"

    def __post_init__(self):
        if self.url is not None and any(
            elem is not None
            for elem in [self.scheme, self.host, self.port, self.params]
        ):
            raise TypeError(
                "If using `url` can not use any of [`scheme`, `host`, `port`, `params`]"
            )


@enforce_types
@dataclass(repr=False, frozen=True)
class Http(BaseResource):
    request: HttpRequest


@enforce_types
@dataclass(repr=False, frozen=True)
class Chain(BaseResource):
    inputs: list


@enforce_types
@dataclass(repr=False, frozen=True)
class Input(BaseResource):
    simple: Optional[dict] = None
    search: Optional[Search] = None
    http: Optional[Http] = None
    chain: Optional[Chain] = None


@enforce_types
@dataclass(repr=False, frozen=True)
class Condition(BaseResource):
    always: Optional[dict] = None
    never: Optional[dict] = None
    compare: Optional[dict] = None
    array_compare: Optional[dict] = None
    script: Optional[dict] = None


@enforce_types
@dataclass(repr=False, frozen=True)
class EmailAction(BaseResource):
    id: str
    account: str
    profile: str
    to: list
    cc: list
    bcc: list
    reply_to: list
    _from: str
    subject: str
    body: str
    body_text: str
    body_html: str
    priority: str
    attachments: str


@enforce_types
@dataclass(repr=False, frozen=True)
class Watch(BaseResource):
    watch_id: str
    trigger: Trigger
    input: Input
    condition: Condition
    actions: dict
    metadata: Optional[dict] = None
    throttle_period: Optional[int] = None
    throttle_period_in_millis: Optional[int] = None

    @property
    def id(self):
        return self.watch_id

    def asdict(self):
        d = super().asdict()
        d.pop("watch_id")
        return d

    @classmethod
    def get(cls, client, watch_id):
        """
        Get a watch from Elasticsearch
        """
        try:
            rt = client.watcher.get_watch(id=watch_id)
        except NotFoundError:
            return None
        schema = WatchSchema()
        try:
            result = schema.load(rt)
        except ValidationError as e:
            raise e
        except TypeError as e:
            raise e
        role = dacite.from_dict(data_class=cls, data=result)
        return role

    def create(self, client):
        try:
            return client.watcher.put_watch(id=self.id, body=self.asdict())
        except Exception as e:
            raise e

    def delete(self, client):
        try:
            return client.watcher.delete_watch(id=self.id)
        except Exception as e:
            raise e

    def update(self, client):
        return self.create(client)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import datetime
import pytz
from dataclasses import FrozenInstanceError
from nemesis.resources.elasticsearch.logstash_pipeline import LogstashPipeline
from elasticsearch import Elasticsearch, RequestError, NotFoundError

from nemesis.tests import es_client, nemesis_client


@pytest.fixture
def pipeline(es_client):
    pipeline = LogstashPipeline(
        id="test-pipeline",
        username="elastic",
        last_modified=datetime.datetime(2022, 3, 3, 0, 0, 0, 0, pytz.UTC),
        pipeline="",
        pipeline_metadata={},
        pipeline_settings={},
    )
    yield pipeline

    # clear remote resources if they exist
    try:
        result = pipeline.delete(es_client)
        assert result == ""
    except NotFoundError:
        pass


def test_asdict(pipeline):
    """ """
    assert pipeline.asdict() == {
        "id": "test-pipeline",
        "last_modified": "2022-03-03T00:00:00.000Z",
        "pipeline": "",
        "pipeline_metadata": {},
        "pipeline_settings": {},
        "username": "elastic",
    }


def test_frozen_object(pipeline):
    with pytest.raises(FrozenInstanceError):
        pipeline.pipeline = "test"


def test_id(pipeline):
    assert pipeline.id == "test-pipeline"


def test_get_not_found(es_client):
    assert LogstashPipeline.get(es_client, "notfound") is None


def test_create_multiple(es_client, pipeline):
    result = pipeline.create(es_client)
    assert result == ""

    result = pipeline.create(es_client)
    assert result == ""


def test_create_update_delete(pipeline, es_client):
    result = pipeline.create(es_client)
    assert result == ""

    result = LogstashPipeline.get(es_client, pipeline.id)
    assert result == pipeline

    # Test that create called multiple times is ok.
    result = pipeline.create(es_client)
    assert result == ""

    new_pipeline = LogstashPipeline(
        id="test-pipeline",
        username="elastic",
        last_modified=datetime.datetime(2022, 3, 14),
        pipeline="test",
        pipeline_metadata={},
        pipeline_settings={},
    )
    result = new_pipeline.update(es_client)
    assert result == ""

    result = pipeline.delete(es_client)
    assert result == ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from marshmallow import Schema, fields, pre_load
from nemesis.schemas.elasticsearch.querydsl import QueryDSLSchema
from nemesis.schemas.elasticsearch.alias import AliasSchema
from nemesis.schemas.elasticsearch.index import IndexSettingsSchema


class TemplateSchema(Schema):
    settings = fields.Nested(IndexSettingsSchema(), data_key="settings")
    mappings = fields.Dict()
    aliases = fields.Nested(AliasSchema(), data_key="aliases")


class IndexTemplateSchema(Schema):
    index_patterns = fields.List(fields.Str())
    template = fields.Nested(TemplateSchema())
    version = fields.Int()
    priority = fields.Int()
    composed_of = fields.List(fields.Str(), allow_none=True)
    version = fields.Int()
    meta = fields.Dict(data_key="_meta")
    name = fields.Str()

    def _remove_empty_values(self, data):
        for k, v in data.items():
            if not v:
                data[k] = None
        return data

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many, **kwargs):
        index_template = data.pop("index_template")
        data.update(index_template)
        data = self._remove_empty_values(data)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import sys
from rich.table import Table
import requests
from elasticsearch import Elasticsearch

from nemesis.console import console
from nemesis.resources import ELASTICSEARCH
from nemesis.utils import Diff


class Nemesis:
    def __init__(self, username, password, es_host=None, cloud_id=None):
        if cloud_id is not None:
            self.client = Elasticsearch(
                cloud_id=cloud_id, http_auth=(username, password)
            )
        else:
            self.client = Elasticsearch(es_host, http_auth=(username, password))
        self.resources = []
        self.diffs = []

    def get_client_for(self, resource):
        if resource.stack_type == ELASTICSEARCH:
            return self.client

    def register(self, resource, force=False, pre_deploy=None, post_deploy=None):
        self.resources.append(
            {
                "resource": resource,
                "remote_resource": resource.get(
                    self.get_client_for(resource), resource.id
                ),
                "force": force,
                "pre_deploy": pre_deploy,
                "post_deploy": post_deploy,
            }
        )

    def preview(self, verbose=False):
        if verbose:
            self._verbose_preview()
        else:
            self._preview()

    def _verbose_preview(self):
        for item in self.resources:
            resource = item["resource"]
            remote_resource = item["remote_resource"]
            self.diffs.append(Diff(resource, remote_resource))

        min_width = max([len(str(x.local_resource)) for x in self.diffs]) + 10

        for diff in self.diffs:
            console.rule()
            table = Table(
                title=f"Preview [resource]{diff.local_resource}",
                min_width=min_width,
            )
            table.add_column("Field", justify="left", no_wrap=True)
            table.add_column("Value", justify="left", no_wrap=False)
            for delta in diff.deltas:
                table.add_row(delta["path"], delta["long"])
            console.print(table)

        created = len([x for x in self.diffs if x.status == "create"])
        updated = len([x for x in self.diffs if x.status == "update"])
        unchanged = len([x for x in self.diffs if x.status == "unchanged"])
        console.print("\nResources:")
        if created > 0:
            console.print(f"\tCreating: {created}")
        if updated > 0:
            console.print(f"\tUpdating: {updated}")
        if unchanged > 0:
            console.print(f"\tUnchanged: {unchanged}")

    def _preview(self):
        table = Table(title="Preview resources to be deployed", show_lines=True)
        table.add_column("Resource", justify="left", no_wrap=True)
        table.add_column("Name", justify="left", no_wrap=True)
        table.add_column("Action", justify="center", no_wrap=True)
        table.add_column("Diff", justify="right", no_wrap=False)
        for item in self.resources:
            resource = item["resource"]
            remote_resource = item["remote_resource"]
            self.diffs.append(Diff(resource, remote_resource))

        for diff in self.diffs:
            local_resource = diff.local_resource
            table.add_row(
                type(local_resource).__name__,
                local_resource.id,
                diff.status,
                "\n".join([x["short"] for x in diff.deltas]),
            )

        created = len([x for x in self.diffs if x.status == "create"])
        updated = len([x for x in self.diffs if x.status == "update"])
        unchanged = len([x for x in self.diffs if x.status == "unchanged"])
        console.print(table)
        console.print("\nResources:")
        if created > 0:
            console.print(f"\tCreating: {created}")
        if updated > 0:
            console.print(f"\tUpdating: {updated}")
        if unchanged > 0:
            console.print(f"\tUnchanged: {unchanged}")

    def _no_change(self, item):
        resource = item["resource"]
        console.print()
        console.rule(f"[resource]{resource}", align="left")
        console.print(
            f"[bold white]No changes detected for [resource]{resource}[/]. Skipping."
        )

    def _pre_deploy(self, resource, function):
        if function is None:
            return
        else:
            console.print()
            console.print(
                f"[bold white]Running pre-deploy function [function]`{function.__name__}`[/]"
            )
            if function(self.client, resource) is False:
                console.print(
                    f"[error]Failed pre_deploy step [function]`{function.__name__}`[/]"
                )
                console.print("[error]Quitting...")
                sys.exit()

    def _post_deploy(self, resource, function):
        if function is None:
            return
        else:
            console.print()
            console.print(
                f"[bold white]Running post-deploy function [function]`{function.__name__}`[/]"
            )
            if function(self.client, resource) is False:
                console.print(
                    f"[error]Failed post_deploy step [function]`{function.__name__}`[/]"
                )
                console.print("[error]Quitting...")
                sys.exit()

    def _force_deploy(self, item):
        resource = item["resource"]
        force = item["force"]
        pre_deploy = item["pre_deploy"]
        post_deploy = item["post_deploy"]
        console.print()
        console.rule(f"[resource]{resource}", align="left")
        console.print(
            f"[bold white]No changes detected for [resource]{resource}[/]. But [add]force[/] applied, so deploying."
        )
        self._pre_deploy(resource, pre_deploy)
        resource.update(self.get_client_for(resource))
        self._post_deploy(resource, post_deploy)

    def _deploy_update(self, item):
        resource = item["resource"]
        force = item["force"]
        pre_deploy = item["pre_deploy"]
        post_deploy = item["post_deploy"]
        console.print()
        console.rule(f"[resource]{resource}", align="left")
        console.print(f"[bold white]Updating resource: [resource]{resource}[/]")
        self._pre_deploy(resource, pre_deploy)
        resource.update(self.get_client_for(resource))
        self._post_deploy(resource, post_deploy)

    def _deploy_create(self, item):
        resource = item["resource"]
        force = item["force"]
        pre_deploy = item["pre_deploy"]
        post_deploy = item["post_deploy"]
        console.print()
        console.rule(f"[resource]{resource}", align="left")
        console.print(f"[bold white]Creating resource: [resource]{resource}[/]")
        self._pre_deploy(resource, pre_deploy)
        resource.create(self.get_client_for(resource))
        self._post_deploy(resource, post_deploy)

    def launch(self, verbose=False):
        console.print()
        console.rule()
        console.rule(f"Preparing to deploy {len(self.resources)} resources...")
        console.rule()
        for item in self.resources:
            resource = item["resource"]
            remote_resource = item["remote_resource"]
            force = item["force"]
            diff = next(
                (x for x in self.diffs if x.local_resource.id == resource.id), None
            )

            if not diff and not force:
                self._no_change(item)
            if not diff and force:
                self._force_deploy(item)
            if diff and remote_resource is not None:
                self._deploy_update(item)
            if diff and remote_resource is None:
                self._deploy_create(item)

        console.rule()
        console.rule(f"Completed deploying {len(self.resources)} resources.")
        console.rule()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import os
import importlib.resources
import click

from nemesis import Nemesis


def load_nemesis_module():
    path = os.getcwd()
    if not os.path.exists(os.path.join(path, "__nemesis__.py")):
        return False
    sys.path.insert(0, path)
    return True


@click.group(chain=True)
def cli():
    """
    Nemesis CLI
    """
    pass


@cli.command("new", help="Create a new nemesis deployment")
def new():
    filename = "__nemesis__.py"
    text = importlib.resources.read_text("nemesis.templates", "__nemesis__.py.tmpl")
    click.echo("Create a new nemesis deployment")
    if not load_nemesis_module():
        if not os.path.exists(filename):
            with open(filename, "w") as f:
                f.write(text)
    else:
        click.echo("Nemesis module already exists.")


@cli.command("preview", help="Preview the resources to deployed")
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="More detailed preview of resources to be deployed.",
)
def preview(verbose):
    if not load_nemesis_module():
        print("No nemesis deployments found.")
        sys.exit()
    try:
        from __nemesis__ import n
    except ImportError as e:
        print("Unable to load nemesis launcher")
        raise
    n.preview(verbose)


@cli.command("launch", help="Deploy resources to Elasticsearch")
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="More detailed preview of resources to be deployed.",
)
@click.option("--autoyes", "-y", is_flag=True, help="Don't prompt before launching")
@click.option("--no-preview", "-p", is_flag=True, help="Don't preview before launching")
def launch(verbose, autoyes, no_preview):
    if not load_nemesis_module():
        print("No nemesis deployments found.")
        sys.exit()
    try:
        from __nemesis__ import n
    except ImportError:
        print("Unable to load nemesis launcher")

    if not no_preview:
        n.preview(verbose)

    if not autoyes:
        confirm = input("Do you want to continue? [y/n default: n]")
    else:
        confirm = "y"

    if confirm.lower() == "y":
        n.launch(verbose)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import inspect
import typing
import dacite
from deepdiff import DeepDiff
from contextlib import suppress
from functools import wraps
from dataclasses import dataclass, asdict

ELASTICSEARCH = "elasticsearch"
KIBANA = "kibana"
RESOURCE_TYPES = (ELASTICSEARCH, KIBANA)


def enforce_types(callable):
    spec = inspect.getfullargspec(callable)

    def check_types(*args, **kwargs):
        parameters = dict(zip(spec.args, args))
        parameters.update(kwargs)
        for name, value in parameters.items():
            with suppress(KeyError):  # Assume un-annotated parameters can be any type
                type_hint = spec.annotations[name]
                if isinstance(type_hint, typing._SpecialForm):
                    # No check for typing.Any, typing.Union, typing.ClassVar (without parameters)
                    continue
                try:
                    actual_type = type_hint.__origin__
                except AttributeError:
                    # In case of non-typing types (such as <class 'int'>, for instance)
                    actual_type = type_hint
                # In Python 3.8 one would replace the try/except with
                # actual_type = typing.get_origin(type_hint) or type_hint
                if isinstance(actual_type, typing._SpecialForm):
                    # case of typing.Union[…] or typing.ClassVar[…]
                    actual_type = type_hint.__args__
                try:
                    if not isinstance(value, actual_type):
                        raise TypeError(
                            "Unexpected type for '{}' (expected {} but found {})".format(
                                name, type_hint, type(value)
                            )
                        )
                except TypeError as e:
                    raise TypeError(
                        f"{e}\n arg1(value)={value}, arg2(actual_type)={actual_type}"
                    )

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            check_types(*args, **kwargs)
            return func(*args, **kwargs)

        return wrapper

    if inspect.isclass(callable):
        callable.__init__ = decorate(callable.__init__)
        return callable

    return decorate(callable)


@dataclass(frozen=True)
class BaseResource:
    def asdict(self):
        return asdict(
            self,
            dict_factory=lambda x: {k: v for (k, v) in x if v is not None},
        )

    def to_json(self):
        return json.dumps(self.asdict(), indent=4)

    @classmethod
    def fromdict(cls, schemaclass, body):
        schema = schemaclass()
        result = schema.load(body)
        return dacite.from_dict(data_class=cls, data=result)

    @property
    def stack_type(self):
        module = self.__module__
        for _type in RESOURCE_TYPES:
            if _type in module:
                return _type

    def diff(self, resource):
        try:
            obj = resource.asdict()
        except AttributeError:
            obj = {}

        return DeepDiff(self.asdict(), obj, view="tree")

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.id})"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from marshmallow import Schema, fields
from nemesis.schemas.elasticsearch.querydsl import QueryDSLSchema


class AliasSchema(Schema):
    filter = fields.Nested(QueryDSLSchema(), data_key="filter")
    index_routing = fields.Str()
    is_hidden = fields.Bool()
    is_write_index = fields.Bool()
    routing = fields.Str()
    search_routing = fields.Str()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from marshmallow import Schema, fields, pre_load, EXCLUDE
from nemesis.schemas.elasticsearch.querydsl import QueryDSLSchema


class Trigger(Schema):
    schedule = fields.Dict()


class Body(Schema):
    query = fields.Nested(QueryDSLSchema)
    size = fields.Int()
    sort = fields.Dict()


class SearchTemplate(Schema):
    id = fields.Str()
    params = fields.Dict()


class SearchRequest(Schema):
    indices = fields.List(fields.Str())
    body = fields.Nested(Body)
    template = fields.Nested(SearchTemplate)


class Search(Schema):
    request = fields.Nested(SearchRequest)
    extract = fields.List(fields.Str())


class HttpRequest(Schema):
    scheme = fields.Str()
    host = fields.Str()
    port = fields.Int()
    path = fields.Str()
    url = fields.Str()
    method = fields.Str()
    body = fields.Str()
    params = fields.Dict()
    headers = fields.Dict()
    auth = fields.Dict()
    proxy = fields.Dict()
    connection_timeout = fields.Str()
    read_timeout = fields.Str()
    extract = fields.List(fields.Str())
    response_content_type = fields.Str()


class Http(Schema):
    request = fields.Nested(HttpRequest)


class Chain(Schema):
    inputs = fields.List(fields.Dict())


class Input(Schema):
    simple = fields.Dict()
    search = fields.Nested(Search)
    http = fields.Nested(Http)
    chain = fields.Nested(Chain)


class Condition(Schema):
    always = fields.Dict()
    never = fields.Dict()
    compare = fields.Dict()
    array_compare = fields.Dict()
    script = fields.Dict()


class EmailAction(Schema):
    id = fields.Str()


class WatchSchema(Schema):
    watch_id = fields.Str()
    trigger = fields.Dict()
    input = fields.Dict()
    condition = fields.Dict()
    actions = fields.Dict()
    metadata = fields.Dict()
    throttle_period = fields.Int()
    throttle_period_in_millis = fields.Int()

    class Meta:
        unknown = EXCLUDE

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many, **kwargs):
        watch = data["watch"]
        watch_id = data["_id"]
        watch["watch_id"] = watch_id
        return watch



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from dataclasses import FrozenInstanceError
from nemesis.resources.elasticsearch.ingest_pipeline import IngestPipeline
from elasticsearch import Elasticsearch, RequestError, NotFoundError

from nemesis.tests import es_client, nemesis_client


@pytest.fixture
def pipeline(es_client):
    pipeline = IngestPipeline(
        id="test-pipeline",
        processors=[
            {"pipeline": {"name": "pipelineA"}},
            {"set": {"field": "outer_pipeline_set", "value": "outer"}},
        ],
    )
    yield pipeline

    # clear remote resources if they exist
    try:
        result = pipeline.delete(es_client)
        assert result == {"acknowledged": True}
    except NotFoundError:
        pass


def test_asdict(pipeline):
    """ """
    assert pipeline.asdict() == {
        "id": "test-pipeline",
        "processors": [
            {"pipeline": {"name": "pipelineA"}},
            {"set": {"field": "outer_pipeline_set", "value": "outer"}},
        ],
    }


def test_frozen_object(pipeline):
    with pytest.raises(FrozenInstanceError):
        pipeline.processors = [{}]


def test_id(pipeline):
    assert pipeline.id == "test-pipeline"


def test_get_not_found(es_client):
    assert IngestPipeline.get(es_client, "notfound") is None


def test_create_multiple(es_client, pipeline):
    result = pipeline.create(es_client)
    assert result == {"acknowledged": True}

    result = pipeline.create(es_client)
    assert result == {"acknowledged": True}


def test_create_update_delete(pipeline, es_client):
    result = pipeline.create(es_client)
    assert result == {"acknowledged": True}

    result = IngestPipeline.get(es_client, pipeline.id)
    assert result == pipeline

    # Test that create called multiple times is ok.
    result = pipeline.create(es_client)
    assert result == {"acknowledged": True}

    pipeline.processors.append(
        {"set": {"field": "inner_pipeline_set", "value": "inner"}}
    )
    result = pipeline.update(es_client)
    assert result == {"acknowledged": True}

    result = pipeline.delete(es_client)
    assert result == {"acknowledged": True}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dataclasses import dataclass, field
from typing import Optional

from nemesis.resources import enforce_types, BaseResource
from nemesis.resources.elasticsearch.querydsl import QueryDSL
from nemesis.resources.elasticsearch.alias import Alias

from nemesis.schemas.elasticsearch.index import IndexSchema

from elasticsearch import RequestError, NotFoundError


@enforce_types
@dataclass(frozen=True)
class IndexSettings(BaseResource):
    index: Optional[dict] = None


@enforce_types
@dataclass(frozen=True)
class Index(BaseResource):
    name: str
    aliases: Optional[Alias] = None
    mappings: Optional[dict] = None
    settings: Optional[IndexSettings] = None

    @property
    def id(self):
        return self.name

    def asdict(self):
        d = super().asdict()
        d.pop("name")
        return d

    @classmethod
    def get(cls, client, name):
        """
        Get an index from Elasticsearch
        """
        try:
            rt = client.indices.get(index=name)
        except NotFoundError:
            return None
        ret = cls.fromdict(schemaclass=IndexSchema, body=rt)
        return ret

    def create(self, client, defer_validation=False, *args, **kwargs):
        try:
            return client.indices.create(
                index=self.id,
                mappings=self.asdict().get("mappings"),
                settings=self.asdict().get("settings"),
                aliases=self.asdict().get("aliases"),
                *args,
                **kwargs,
            )
        except RequestError as e:
            raise e

    def delete(self, client):
        try:
            return client.indices.delete(index=self.id)
        except RequestError as e:
            raise e



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nemesis',
 'nemesis.resources',
 'nemesis.resources.elasticsearch',
 'nemesis.schemas',
 'nemesis.schemas.elasticsearch',
 'nemesis.scripts',
 'nemesis.templates',
 'nemesis.tests',
 'nemesis.tests.resources.elasticsearch']

package_data = \
{'': ['*']}

install_requires = \
['Click>=7.1,<7.2',
 'dacite>=1.6.0,<2.0.0',
 'deepdiff>=5.6.0,<6.0.0',
 'elasticsearch>=7.15.2,<8.0.0',
 'marshmallow>=3.14.1,<4.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'pytz>=2021.3,<2022.0',
 'requests>=2.26.0,<3.0.0',
 'rich>=10.14.0,<11.0.0']

entry_points = \
{'console_scripts': ['nemesis = nemesis.scripts.nemesis:cli']}

setup_kwargs = {
    'name': 'nemesis',
    'version': '0.0.8',
    'description': 'Tool for managing Elasticsearch resources as code',
    'long_description': '# Nemesis ![tests-badge](https://github.com/fxdgear/nemesis/actions/workflows/python-app.yml/badge.svg)\n\nNemesis is a python library to manage Elasticsearch resources as code. Nemesis operates\na lot more like Pulumi than terraform. Each resource that nemesis supports is an actual\npython object which can be used like any other python object.\n\nElasticsearch resources can be crafted as Python objects.\nElasticsearch resources can be fetched from the Elasticsearch cluster and diffed against local versions.\nDeployments can happen if a remote resource doesn\'t exist.\nDeployments can happen if a local_resource is registered with force=True, to force updating of the\nresource even if it hasn\'t changed.\n\nCreating new resources is not trivial at this point in time.\nIt\'s difficult because, to create the resource you first need to define the dataclass and all the attributes on that class.\nNext you would need to define the Schema for that resource, then you would need to define all the CRUD method for that resource.\nThere\'s no automated way to do this. But would be nice to create a code generator to scan the Elasticsearch repo and pull out the various resources and their types so they can be created in Nemesis.\n\nPull requests for new resources added to nemesis would be greatly appriciated!\n\n## Getting started.\n\n### Installation\nFirst to get started you must install `nemesis`:\n\n```\npip install nemesis\n```\n\n### Creating your first nemesis project\n\nFirst create the directory you want to put your nemesis project into:\n\n```\n$ mkdir my_first_project\n$ cd my_first_project\n$ nemesis new\n```\n\n### The __nemesis__.py file\n\nAfter you run the `nemesis new` command a newly created __nemesis__.py file will exist. This file has some example code in it to help you get started.\n\n1. instantiate the `Nemesis` object as the variable `n`.\n2. using the `nemesis.resources.elasticsearch.*` modules create your ES resources\n3. `register` those resources with the `Nemesis` client. `n.register(my_resource_name)`\n\n### Help\n\nRunning the `--help` flag will give you help on any command or subcommand.\n\n```\n> nemesis --help\nUsage: nemesis [OPTIONS] COMMAND1 [ARGS]... [COMMAND2 [ARGS]...]...\n\n  Nemesis CLI\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  launch   Deploy resources to Elasticsearch\n  new      Create a new nemesis deployment\n  preview  Preview the resources to deployed\n```\n\n\n### Preview\n\nRun `nemesis preview` to see what will be deployed to Elasticsearch.\n\nThis will render a diff to tell you what\'s going to be created or changed.\n\n### Launch\n\nRun `nemesis launch` to deploy your changes.\n\nThis will actually ship your resources to Elasticsearch\n\n---\n\n# Additional features\n\n## Pre/Post deploy hooks.\nnemesis supports pre and post deploy hooks. This is useful in various situations:\n\n1. You are creating an ingest pipeline and you want to run some tests to ensure your pipeline works before the pipeline is deployed.\nYou can write a function to call `resource.simulate` on  your pipeline resource.\n2. You have a transform you want to "reset". You can define a "predeploy" function to "stop" the current transform,\n  delete the current dest index, and then recreate the dest index. then define a "postdeploy" function which will start the transform\n3. Watchers can be "simulated" using the "execute" api. This unfortunatly only works on watchers that exist already, but you can deploy the watcher and then write a post-deploy\ntest that will ensure your watcher works as expected.\n4. Your imagination is the limit on things you might want to do before and/or after something has been deployed\n\n---\n\n# Diffing elasticsearch resources\n\nNemesis supports deep diffing of resources with a succinct mode and verbose mode.\n\nExample of a "succinct" diff:\n\n```\n> nemesis launch -y\nHello Nemesis!\n                             Preview resources to be deployed\n┏━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┓\n┃ Resource         ┃ Name                             ┃  Action   ┃                 Diff ┃\n┡━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━┩\n│ IndexTemplate    │ index_template_billing_aggregate │  create   │ + [\'index_patterns\'] │\n│                  │                                  │           │       + [\'template\'] │\n│                  │                                  │           │        + [\'version\'] │\n├──────────────────┼──────────────────────────────────┼───────────┼──────────────────────┤\n│ IngestPipeline   │ total_cost                       │  create   │             + [\'id\'] │\n│                  │                                  │           │     + [\'processors\'] │\n│                  │                                  │           │    + [\'description\'] │\n│                  │                                  │           │        + [\'version\'] │\n├──────────────────┼──────────────────────────────────┼───────────┼──────────────────────┤\n│ Transform        │ total_cost_2021_12               │  create   │         + [\'source\'] │\n│                  │                                  │           │           + [\'dest\'] │\n│                  │                                  │           │             + [\'id\'] │\n│                  │                                  │           │          + [\'pivot\'] │\n│                  │                                  │           │    + [\'description\'] │\n│                  │                                  │           │      + \'frequency\'] │\n├──────────────────┼──────────────────────────────────┼───────────┼──────────────────────┤\n│ LogstashPipeline │ test_logstash_pipeline           │  update   │  ~ [\'last_modified\'] │\n├──────────────────┼──────────────────────────────────┼───────────┼──────────────────────┤\n│ Role             │ test-role                        │ unchanged │                      │\n├──────────────────┼──────────────────────────────────┼───────────┼──────────────────────┤\n│ RoleMapping      │ test_role_mapping                │ unchanged │                      │\n├──────────────────┼──────────────────────────────────┼───────────┼──────────────────────┤\n│ Watch            │ test-watch                       │ unchanged │                      │\n└──────────────────┴──────────────────────────────────┴───────────┴──────────────────────┘\n\nResources:\n        Creating: 3\n        Updating: 1\n        Unchanged: 3\n```\n\nAnd an example of a verbose diff:\n```\n> nemesis launch -v\nHello Nemesis!\n────────────────────────────────────────────────────────────────────────────────────────────────────────\n Preview IndexTemplate(index_template_billing_aggregate)\n┏━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓\n┃ Field               ┃ Value                           ┃\n┡━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩\n│ [\'index_patterns\']  │ + [                             │\n│                     │   "billing_aggregate_*"         │\n│                     │ ]                               │\n│ [\'template\']        │ + {                             │\n│                     │   "settings": {                 │\n│                     │     "index": {                  │\n│                     │       "number_of_shards": "1",  │\n│                     │       "number_of_replicas": "1" │\n│                     │     }                           │\n│                     │   },                            │\n│                     │   "mappings": {                 │\n│                     │     "_source": {                │\n│                     │       "enabled": true           │\n│                     │     },                          │\n│                     │     "properties": {             │\n│                     │       "@timestamp": {           │\n│                     │         "type": "date"          │\n│                     │       },                        │\n│                     │       "@version": {             │\n│                     │         "type": "text"          │\n│                     │       },                        │\n│                     │       "cloud_provider": {       │\n│                     │         "type": "keyword"       │\n│                     │       },                        │\n│                     │       "sum_total": {            │\n│                     │         "type": "float"         │\n│                     │       },                        │\n│                     │       "team_name": {            │\n│                     │         "type": "keyword"       │\n│                     │       }                         │\n│                     │     }                           │\n│                     │   }                             │\n│                     │ }                               │\n│ [\'version\']         │ + 3                             │\n└─────────────────────┴─────────────────────────────────┘\n```\n',
    'author': 'Nick Lang',
    'author_email': 'nick@nicklang.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fxdgear/nemesis',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime
from elasticsearch import NotFoundError
import dacite
from dataclasses import dataclass, field
from typing import Optional
from nemesis.resources import enforce_types, BaseResource
from nemesis.schemas.elasticsearch.logstash_pipeline import LogstashPipelineSchema


def time_format(dt):
    """
    timeformat must match Elasticsearch `strict_date_time` format:
    https://www.elastic.co/guide/en/elasticsearch/reference/7.16/mapping-date-format.html
    """
    s = dt.strftime("%Y-%m-%dT%H:%M:%S.%f")
    return f"{s[:-3]}Z"


@enforce_types
@dataclass(repr=False, frozen=True)
class LogstashPipeline(BaseResource):
    id: str
    last_modified: datetime
    pipeline: str
    pipeline_metadata: dict
    pipeline_settings: dict
    username: str
    description: Optional[str] = None

    def asdict(self):
        d = super().asdict()
        d["last_modified"] = time_format(d["last_modified"])
        return d

    @classmethod
    def get(cls, client, pipeline_id):
        """
        Get a logstash pipeline from Elasticsearch
        """
        try:
            pipeline = client.logstash.get_pipeline(id=pipeline_id)
        except NotFoundError:
            return None
        schema = LogstashPipelineSchema()
        result = schema.load(pipeline)
        pipeline = dacite.from_dict(data_class=cls, data=result)
        return pipeline

    def create(self, client):
        body = self.asdict()
        body.pop("id")
        return client.logstash.put_pipeline(id=self.id, body=body)

    def delete(self, client):
        try:
            return client.logstash.delete_pipeline(id=self.id)
        except Exception as e:
            raise e

    def update(self, client):
        return self.create(client)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dataclasses import dataclass, field
from typing import Optional

from nemesis.resources import enforce_types, BaseResource
from nemesis.resources.elasticsearch.querydsl import QueryDSL


@enforce_types
@dataclass(frozen=True)
class Alias(BaseResource):
    filter: Optional[QueryDSL]
    index_routing: Optional[str]
    is_hidden: Optional[bool]
    is_write_index: Optional[bool]
    routing: Optional[str]
    search_routing: Optional[str]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import dacite
from marshmallow.exceptions import ValidationError
from dataclasses import dataclass, field
from typing import Optional, Union
from nemesis.schemas.elasticsearch.security import RoleSchema, RoleMappingSchema
from nemesis.resources import enforce_types, BaseResource
from nemesis.resources.elasticsearch.querydsl import QueryDSL


@enforce_types
@dataclass(frozen=True)
class Application(BaseResource):
    application: str
    privileges: list
    resources: list


@enforce_types
@dataclass(frozen=True)
class Index(BaseResource):
    names: list
    privileges: list
    query: Optional[QueryDSL] = None
    field_security: Optional[dict] = None
    allow_restricted_indices: Optional[bool] = None


@enforce_types
@dataclass(repr=False, frozen=True)
class Role(BaseResource):
    name: str
    applications: list[Application]
    cluster: list
    indices: list[Index]
    metadata: dict
    run_as: Optional[list] = None
    _global: Optional[dict] = None

    @property
    def id(self):
        return self.name

    def asdict(self):
        """
        The "name" field isn't part of the actual body sent to Elasticsearch.
        But it's nice to have on the object we are dealing with.
        """
        d = super().asdict()
        d.pop("name")
        return d

    @classmethod
    def get(cls, client, name):
        """
        Get a role from Elasticsearch
        """
        rt = client.security.get_role(name=name)
        schema = RoleSchema()
        try:
            result = schema.load(rt)
        except ValidationError as e:
            raise e
        role = dacite.from_dict(data_class=cls, data=result)
        return role

    def create(self, client):
        body = self.asdict()
        try:
            return client.security.put_role(name=self.id, body=body)
        except Exception as e:
            raise e

    def update(self, client):
        return self.create(client)

    def delete(self, client):
        try:
            return client.security.delete_role(name=self.id)
        except Exception as e:
            raise e


@enforce_types
@dataclass(repr=False, frozen=True)
class RoleMapping(BaseResource):
    name: str
    enabled: bool
    rules: dict
    roles: Optional[list] = None
    role_templates: Optional[list] = None
    metadata: Optional[dict] = None

    def asdict(self):
        """
        The "name" field isn't part of the actual body sent to Elasticsearch.
        But it's nice to have on the object we are dealing with.
        """
        d = super().asdict()
        d.pop("name")
        return d

    @property
    def id(self):
        return self.name

    def __post_init__(self):
        if self.roles is None and self.role_templates is None:
            raise TypeError(
                "Value needed for one of either `roles` or `role_templates`."
            )

    @classmethod
    def get(cls, client, name):
        """
        Get a role from Elasticsearch
        """
        rt = client.security.get_role_mapping(name=name)
        schema = RoleMappingSchema()
        try:
            result = schema.load(rt)
        except ValidationError as e:
            raise e
        role = dacite.from_dict(data_class=cls, data=result)
        return role

    def create(self, client):
        body = self.asdict()
        try:
            return client.security.put_role_mapping(name=self.id, body=body)
        except Exception as e:
            raise e

    def update(self, client):
        return self.create(client)

    def delete(self, client):
        body = self.asdict()
        try:
            return client.security.delete_role_mapping(name=self.id)
        except Exception as e:
            raise e



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from dataclasses import FrozenInstanceError
from nemesis.resources.elasticsearch.security import (
    Role,
    RoleMapping,
    Index,
    Application,
)
from nemesis.resources.elasticsearch.querydsl import QueryDSL
from elasticsearch import Elasticsearch, RequestError

from nemesis.tests import es_client, nemesis_client


@pytest.fixture
def role():
    return Role(
        name="test-role",
        cluster=["all"],
        indices=[
            Index(
                names=["index1", "index2"],
                privileges=["all"],
                field_security={"grant": ["title", "body"]},
                query=QueryDSL(match={"title": "foo"}),
            )
        ],
        applications=[
            Application(
                application="myapp", privileges=["admin", "read"], resources=["*"]
            )
        ],
        run_as=["other_user"],
        metadata={"version": "1"},
    )


@pytest.fixture
def role_mapping(role):
    return RoleMapping(
        name="test-role-mapping",
        enabled=True,
        rules={"field": {"username": "*"}},
        roles=[role.name],
    )


def test_asdict(role, role_mapping):
    """
    asdict method pop's the `name` field from the object to match
    how it would be sent or received from Elasticsearch
    """
    assert role.asdict() == {
        "applications": [
            {
                "application": "myapp",
                "privileges": ["admin", "read"],
                "resources": ["*"],
            }
        ],
        "cluster": ["all"],
        "indices": [
            {
                "names": ["index1", "index2"],
                "privileges": ["all"],
                "query": {"match": {"title": "foo"}},
                "field_security": {"grant": ["title", "body"]},
            }
        ],
        "metadata": {"version": "1"},
        "run_as": ["other_user"],
    }

    assert role_mapping.asdict() == {
        "enabled": True,
        "rules": {"field": {"username": "*"}},
        "roles": [
            "test-role",
        ],
    }


def test_frozen_object(role, role_mapping):
    with pytest.raises(FrozenInstanceError):
        role.cluster = ["raise-error"]

    with pytest.raises(FrozenInstanceError):
        role_mapping.enabled = False


def test_id(role, role_mapping):
    assert role.id == "test-role"
    assert role_mapping.id == "test-role-mapping"


def test_create_update_delete(role, role_mapping, es_client):
    result = role.create(es_client)
    assert result == {"role": {"created": True}}

    result = role_mapping.create(es_client)
    assert result == {"role_mapping": {"created": True}}

    # test what happens when we call create a 2nd time
    result = role.create(es_client)
    assert result == {"role": {"created": False}}

    result = role_mapping.create(es_client)
    assert result == {"role_mapping": {"created": False}}

    result = role.update(es_client)
    assert result == {"role": {"created": False}}

    result = role_mapping.update(es_client)
    assert result == {"role_mapping": {"created": False}}

    result = role.delete(es_client)
    assert result == {"found": True}

    result = role_mapping.delete(es_client)
    assert result == {"found": True}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
from marshmallow import Schema, fields, pre_load
from nemesis.schemas.elasticsearch.querydsl import QueryDSLSchema
from nemesis.schemas.elasticsearch.index import IndexSchema
from nemesis.schemas.elasticsearch.ingest_pipeline import IngestPipelineSchema


class SyncTimeSchema(Schema):
    field = fields.Str()
    delay = fields.Str(dump_default="60s")


class SyncSchema(Schema):
    time = fields.Nested(SyncTimeSchema)


class RetentionPolicyTimeSchema(Schema):
    field = fields.Str()
    max_age = fields.Str()


class RetentionPolicySchema(Schema):
    time = fields.Nested(RetentionPolicyTimeSchema, data_key="time")


class SettingsSchema(Schema):
    dates_as_epoc_millis = fields.Boolean(dump_default=False)
    docs_per_second = fields.Float()
    align_checkpoints = fields.Boolean(dump_default=True)
    max_page_search_size = fields.Integer(dump_default=500)


class LatestSchema(Schema):
    sort = fields.Str()
    unique_key = fields.List(fields.Str())


class PivotSchema(Schema):
    aggregations = fields.Dict()
    group_by = fields.Dict()


class DestSchema(Schema):
    index = fields.Str()
    pipeline = fields.Str()


class SourceSchema(Schema):
    index = fields.List(fields.Str())
    runtime_mappings = fields.Dict()
    query = fields.Nested(QueryDSLSchema(), data_key="query")


class TransformSchema(Schema):
    # fmt: off
    source = fields.Nested(SourceSchema, data_key="source")
    dest = fields.Nested(DestSchema, data_key="dest")
    pivot = fields.Nested(PivotSchema, data_key="pivot")
    latest = fields.Nested(LatestSchema, data_key="latest")
    sync = fields.Nested(SyncSchema, data_key="sync")
    retention_policy = fields.Nested(RetentionPolicySchema, data_key="retention_policy")
    settings = fields.Nested(SettingsSchema, data_key="settings", allow_none=True)
    description = fields.Str()
    frequency = fields.Str()
    version = fields.Str()
    create_time = fields.Int()
    version = fields.Str()
    id = fields.Str()
    # fmt: on

    @pre_load
    def remove_empty_values(self, data, *args, **kwargs):
        for k, v in data.items():
            if not v:
                data[k] = None
        return data
